# coding: utf-8
# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-celerytask automatic tests"""

import collections
import datetime
import json
import logging
import time
import os.path

import six
import celery
import celery.result
import mock

from cubicweb.devtools import testlib  # noqa

from cubicweb_celerytask.ccplugin import CeleryMonitorCommand
from cubicweb_celerytask.entities import (start_async_task, StartCeleryTaskOp,
                                          run_all_tasks)
from cubicweb_celerytask.testutils import BaseCeleryTaskTC

from cw_celerytask_helpers.monitor import MONITOR_KEY
from cw_celerytask_helpers.utils import get_redis_client
from cw_celerytask_helpers.filelogger import get_log_filename


def wait_until(func, timeout=10, retry=1):
    start = time.time()
    while abs(time.time() - start) < timeout:
        if func():
            break
        time.sleep(retry)
    else:
        raise AssertionError(
            'predicate has not been verified since %ds' % timeout)


class CeleryTaskTC(BaseCeleryTaskTC):

    def test_success_task(self):
        with self.admin_access.repo_cnx() as cnx:
            cwtask_eid = start_async_task(cnx, 'success', 42).eid
            cnx.commit()
            cwtask = cnx.entity_from_eid(cwtask_eid)
            run_all_tasks(cnx)
            self.wait_async_task(cnx, cwtask.task_id)
            result = cwtask.cw_adapt_to('ICeleryTask').result
            self.assertEqual(result.get(), 42)
            wf = cwtask.cw_adapt_to('IWorkflowable')
            self.assertEqual(wf.state, 'done')

    def test_fail_task(self):
        with self.admin_access.repo_cnx() as cnx:
            cwtask_eid = start_async_task(cnx, 'fail').eid
            cnx.commit()
            run_all_tasks(cnx)
            cwtask = cnx.entity_from_eid(cwtask_eid)
            self.wait_async_task(cnx, cwtask.task_id)
            result = cwtask.cw_adapt_to('ICeleryTask').result
            tb = result.traceback
            self.assertTrue(result.failed())
            self.assertTrue(tb.startswith('Traceback (most recent call last)'))
            self.assertTrue(tb.endswith('RuntimeError: fail\n'))
            wf = cwtask.cw_adapt_to('IWorkflowable')
            self.assertEqual(wf.state, 'failed')

    def test_exc_encoding(self):
        with self.admin_access.repo_cnx() as cnx:
            cwtask_eid = start_async_task(cnx, 'exc_encoding').eid
            cnx.commit()
            run_all_tasks(cnx)
            cwtask = cnx.entity_from_eid(cwtask_eid)
            self.wait_async_task(cnx, cwtask.task_id)
            wf = cwtask.cw_adapt_to('IWorkflowable')
            self.assertEqual(wf.state, 'failed')
            logs = cwtask.cw_adapt_to('ICeleryTask').logs.decode('utf-8')
            if six.PY2:
                self.assertIn(
                    (u"""raise RuntimeError(u'Cette tâche a échoué'."""
                     u"""encode('utf8'))"""), logs)
            else:
                self.assertIn(
                    u"""raise RuntimeError('Cette tâche a échoué')""",
                    logs)
            self.assertIn(u"""RuntimeError: Cette tâche a échoué""", logs)

    def test_start_celerytask_op(self):
        with self.admin_access.repo_cnx() as cnx:
            # Task must run even if there is no data on current transaction
            task = celery.signature("success", kwargs={"n": 10})
            task_id = task.freeze().task_id
            cwtask = cnx.create_entity('CeleryTask', task_name=u'dummy',
                                       task_id=six.text_type(task_id))
            cnx.commit()
            cnx.transaction_data['celerytask'] = {cwtask.eid: task}
            StartCeleryTaskOp.get_instance(cnx).add_data(cwtask.eid)
            cnx.commit()
            run_all_tasks(cnx)
            result = self.wait_async_task(cnx, task_id)
            self.assertEqual(result.get(), 10)

    def test_logs(self):
        with self.admin_access.repo_cnx() as cnx:
            cwtask_eid = start_async_task(cnx, 'log').eid
            cnx.commit()
            run_all_tasks(cnx)
            cwtask = cnx.entity_from_eid(cwtask_eid)
            self.wait_async_task(cnx, cwtask.task_id)
            logs = cwtask.cw_adapt_to('ICeleryTask').logs
            self.assertIn(b'out should be in logs', logs)
            self.assertIn(b'err should be in logs', logs)
            self.assertIn(b'cw warning should be in logs', logs)
            for name in (b'cw', b'celery'):
                for key in (b'error', b'critical', b'exception'):
                    self.assertIn(name + b' ' + key + b' should be in logs',
                                  logs)
            self.assertIn(b'cw critical should be in logs', logs)
            self.assertNotIn(b'should not be in logs', logs)
            self.assertIn(b'raise Exception("oops")', logs)

    def test_task_deleted(self):
        rdb = get_redis_client()
        with self.admin_access.cnx() as cnx:
            # this 'buggy_task_revoked' key is used simulate the 'revoke' since
            # it's not handled by celery in threaded solo mode that we use in
            # tests
            rdb.set('buggy_task_revoked', 'no')
            task = start_async_task(cnx, 'buggy_task')
            cnx.commit()
            run_all_tasks(cnx)
            cnx.commit()
            task = cnx.entity_from_eid(task.eid)
            wait_until(lambda: b'evil' in task.cw_adapt_to('ICeleryTask').logs)
            task.cw_delete()
            with mock.patch('celery.app.control.Control.revoke') as revoke:
                cnx.commit()
            rdb.set('buggy_task_revoked', 'yes')
            revoke.assert_called_once_with([task.task_id], signal='SIGKILL',
                                           terminate=True)
            self.assertFalse(os.path.exists(get_log_filename(task.task_id)))

    def test_workflow_chain(self):
        with self.admin_access.repo_cnx() as cnx:
            s = celery.signature
            task = celery.chain(s("add", (2, 2)), s("add", (4,)))
            cwtask = start_async_task(cnx, task)
            cnx.commit()
            run_all_tasks(cnx)
            result = self.wait_async_task(cnx, cwtask.task_id)
            self.assertEqual(result.get(), 8)

            children = cwtask.reverse_parent_task
            self.assertEqual(len(children), 1)
            result = children[0].cw_adapt_to('ICeleryTask').result
            self.assertEqual(result.get(), 4)

    def test_workflow_group(self):
        with self.admin_access.repo_cnx() as cnx:
            s = celery.signature
            task = celery.group(s("add", (2, 3)), s("add", (4, 5)))
            cwtask = start_async_task(cnx, task)
            cnx.commit()
            self.assertEqual(cwtask.task_name, u'celery.group')
            run_all_tasks(cnx)
            # FIXME: investigate why this hang since update from celery 3 to 4
            # results = run_all_tasks(cnx)
            # self.wait_async_task(cnx, cwtask.task_id)
            # self.assertEqual(results[cwtask.eid].get(), [5, 9])
            subtasks = cwtask.reverse_parent_task
            self.assertEqual(len(subtasks), 2)
            for subtask, expected in zip(subtasks, [5, 9]):
                self.wait_async_task(cnx, subtask.task_id)
                result = subtask.cw_adapt_to('ICeleryTask').result.get()
                self.assertEqual(result, expected)

    def test_workflow_chord(self):
        with self.admin_access.repo_cnx() as cnx:
            s = celery.signature
            task = celery.chord([s("success", (i,)) for i in range(10)],
                                s("tsum", []))
            cwtask = start_async_task(cnx, task)
            cnx.commit()
            self.assertEqual(cwtask.task_name, u'celery.chord')
            run_all_tasks(cnx)
            result = self.wait_async_task(cnx, cwtask.task_id)
            self.assertEqual(result.get(), 45)

            children = cwtask.reverse_parent_task
            self.assertEqual([child.task_name for child in children],
                             [u'success'] * 10)
            self.assertCountEqual(
                [t.cw_adapt_to('ICeleryTask').result.get() for t in children],
                range(10))

    def test_workflow_subtasks(self):
        with self.admin_access.repo_cnx() as cnx:
            logging.getLogger('cubicweb.appobject').setLevel(logging.DEBUG)
            s = celery.signature
            task = s("spawn")
            cwtask = start_async_task(cnx, task)
            cnx.commit()
            self.assertEqual(cwtask.task_name, u'spawn')
            run_all_tasks(cnx)
            asresult = self.wait_async_task(cnx, cwtask.task_id)
            result = celery.result.result_from_tuple(
                asresult.result["celerytask_subtasks"])
            self.assertEqual(0, result.get())
            cwtask = cnx.entity_from_eid(cwtask.eid)
            self.assertEqual(cwtask.cw_adapt_to('IWorkflowable').state, "done")
            children = cwtask.reverse_parent_task
            self.assertEqual(len(children), 11)
            counter = collections.Counter()
            states = []
            for child in children:
                counter[child.task_name] += 1
                states.append(child.cw_adapt_to('IWorkflowable').state)
            self.assertEqual(dict(counter), {'success': 10, 'add': 1})

            tsum = cnx.find('CeleryTask', task_name='tsum').one()
            self.assertEqual(tsum.cw_adapt_to('IWorkflowable').state, 'done')
            # XXX: this task should have "spawn" as parent_task ?
            self.assertEqual(tsum.parent_task, ())
            self.assertEqual(tsum.reverse_parent_task, ())

    def test_revoke(self):
        with self.admin_access.cnx() as cnx:
            task = start_async_task(cnx, celery.signature("success"))
            cnx.commit()
            with mock.patch('celery.app.control.Control.revoke') as revoke:
                task.cw_adapt_to('ICeleryTask').revoke()
            revoke.assert_called_once_with([task.task_id], signal='SIGKILL',
                                           terminate=True)

    def test_multi_revoke(self):
        with self.admin_access.cnx() as cnx:

            def success():
                return celery.signature("success")

            task = start_async_task(cnx, celery.group(success(), success()))
            cnx.commit()
            task_ids = set([t.task_id for t in task.child_tasks()])
            # we have a task group containing two subtasks = 3 tasks
            self.assertEqual(len(task_ids), 3)
            with mock.patch('celery.app.control.Control.revoke') as revoke:
                task.cw_adapt_to('ICeleryTask').revoke()
            self.assertEqual(revoke.call_count, 1)
            args, kwargs = revoke.call_args
            self.assertEqual(len(args), 1)
            self.assertEqual(set(task_ids), set(args[0]))
            self.assertEqual({'terminate': True, 'signal': 'SIGKILL'}, kwargs)

    def test_celery_monitor_requeue(self):
        with self.admin_access.cnx() as cnx:
            t1 = start_async_task(cnx, "success", 42)
            t2 = start_async_task(cnx, "success", 42)
            t1.cw_set(creation_date=(
                datetime.datetime.utcnow() - datetime.timedelta(hours=2)))
            cnx.commit()
            rdb = get_redis_client()
            self.assertEqual(rdb.lrange(MONITOR_KEY, 0, -1), [])
            CeleryMonitorCommand.requeue(cnx, rdb)
            self.assertEqual(
                [json.loads(data.decode())
                 for data in rdb.lrange(MONITOR_KEY, 0, -1)],
                [{'task_id': t2.task_id, 'task_name': None}])


class StartAsyncTaskTC(testlib.CubicWebTC):

    def setUp(self):
        super(StartAsyncTaskTC, self).setUp()
        celery.current_app.conf.task_always_eager = True

    def test_task_creating_task(self):
        with self.admin_access.cnx() as cnx:

            @celery.current_app.task(name='task_a')
            def task_a():
                with self.admin_access.cnx() as admin_cnx:
                    start_async_task(admin_cnx, 'task_b')
                    admin_cnx.commit()
                return 'a'

            @celery.current_app.task(name='task_b')
            def task_b():
                return 'b'

            @celery.current_app.task(name='task_c')
            def task_c():
                return 'c'

            start_async_task(cnx, 'task_a')
            start_async_task(cnx, 'task_c')
            cnx.commit()
            results = run_all_tasks(cnx)
            self.assertCountEqual(
                [r.get() for r in results.values()], ['a', 'b', 'c'])


class SegfaultTC(BaseCeleryTaskTC):
    worker_args = ['-P', 'prefork', '-c', '1', '-l', 'info']

    def test_segfault(self):
        with self.admin_access.cnx() as cnx:
            task = start_async_task(cnx, 'segfault')
            cnx.commit()
            run_all_tasks(cnx)
            self.wait_async_task(cnx, task.task_id, timeout=15)
            wf = cnx.entity_from_eid(task.eid).cw_adapt_to('IWorkflowable')
            self.assertEqual(wf.state, 'failed')
            self.assertEqual(wf.latest_trinfo().comment, (
                'Worker exited prematurely: signal 11 (SIGSEGV).'))


if __name__ == '__main__':
    from unittest import main
    main()



# coding: utf-8
from __future__ import print_function

import sys
import logging
import os
import time

import six

from celery import current_app as app, chord
from celery.utils.log import get_task_logger

from cw_celerytask_helpers.utils import get_redis_client
from cw_celerytask_helpers import redirect_stdouts

cw_logger = logging.getLogger('cubes.tasks')
dummy_logger = logging.getLogger('dummy')
logger = get_task_logger(__name__)


@app.task(name='success')
def success(n):
    return n


@app.task(name='fail')
def fail():
    raise RuntimeError('fail')


@app.task(bind=True, name='exc_encoding')
def exc_encoding(self):
    if six.PY2:
        raise RuntimeError(u'Cette tâche a échoué'.encode('utf8'))
    else:
        raise RuntimeError('Cette tâche a échoué')


@app.task(name='buggy_task')
def buggy_task():
    rdb = get_redis_client()
    while True:
        if rdb.get('buggy_task_revoked') == b'yes':
            break
        logger.error('evil')
        time.sleep(1)


@app.task(name='log')
def log():
    for out in [sys.stdout, sys.stderr]:
        print('should not be in logs', file=out)

    with redirect_stdouts(logger):
        print('out should be in logs')
        print('err should be in logs', file=sys.stderr)

    for out in [sys.stdout, sys.stderr]:
        print('should not be in logs', file=out)

    for name, l, state in [
        ('cw', cw_logger, 'be'),
        ('celery', logger, 'be'),
        ('dummy', dummy_logger, 'not be')
    ]:
        for key in ('debug', 'info', 'warning', 'error', 'critical'):
            getattr(l, key)('%s %s should %s in logs', name, key, state)
        try:
            raise RuntimeError("fail")
        except RuntimeError:
            l.exception('%s exception should be in logs', name)

    raise Exception("oops")


@app.task(name="add")
def add(x, y):
    return x + y


@app.task(name="tsum")
def tsum(args):
    return sum(args)


@app.task(bind=True, name="spawn")
def spawn(self):
    return {
        "celerytask_subtasks": chord((success.s(i) for i in range(10)),
                                     tsum.s() | add.s(-45))().as_tuple(),
    }

@app.task(name="segfault")
def segfault():
    import ctypes
    ctypes.string_at(0)



# -*- coding: utf-8 -*-
# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-celerytask specific hooks and operations"""
from __future__ import print_function

import errno
import os.path

from celery import current_app
import celery.task.control

from cubicweb import ConfigurationError
from cubicweb.predicates import is_instance
from cubicweb.server.hook import Hook, DataOperationMixIn, Operation

from cw_celerytask_helpers.filelogger import flush_task_logs


class DeleteCeleryTaskOp(DataOperationMixIn, Operation):

    def postcommit_event(self):
        tasks = set(self.get_data())
        if tasks:
            celery.task.control.revoke(list(tasks), terminate=True,
                                       signal='SIGKILL')
        for task_id in tasks:
            flush_task_logs(task_id)


class CeleryTaskDeletedHook(Hook):
    """revoke task and flush task logs when a task is deleted"""
    __regid__ = 'celerytask.celerytask_deleted'
    __select__ = Hook.__select__ & is_instance('CeleryTask')
    events = ('before_delete_entity',)

    def __call__(self):
        op = DeleteCeleryTaskOp.get_instance(self._cw)
        for task in self.entity.child_tasks():
            op.add_data(task.task_id)


class CeleryTaskStartupHook(Hook):
    """Initialize CUBICWEB_CELERYTASK_LOGDIR"""
    __regid__ = 'celerytask.server-startup-hook'
    events = ('server_startup', 'server_maintenance')

    def __call__(self):
        self.setup_celerytask_logdir(self.repo.vreg.config)

    @staticmethod
    def setup_celerytask_logdir(config):
        celery_logdir = current_app.conf.get('CUBICWEB_CELERYTASK_LOGDIR')
        logdir = config['celerytask-log-dir']
        if celery_logdir and logdir and celery_logdir != logdir:
            raise ConfigurationError((
                "You misconfigured your application by setting two different "
                "celerytask log directories: "
                "{} in celeryconfig and {} in cubicweb all-in-one.conf"
            ).format(celery_logdir, logdir))
        elif celery_logdir and not logdir:
            logdir = config['celerytask-log-dir'] = celery_logdir
        elif not celery_logdir:
            if not logdir:
                logdir = os.path.join(config.appdatahome, 'logs')
            current_app.conf['CUBICWEB_CELERYTASK_LOGDIR'] = logdir
            config['celerytask-log-dir'] = logdir
        # ensure directory exist
        try:
            os.makedirs(logdir)
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise



# -*- coding: utf-8 -*-
# copyright 2018 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import gzip

from logilab.common.shellutils import ProgressBar
from cw_celerytask_helpers.filelogger import get_log_filename
from cw_celerytask_helpers.redislogger import get_task_logs, flush_task_logs


def migrate_task_logs_to_bfss(cnx):
    """Migrate logs from redis and from database to logs files in
    celerytask-log-dir"""
    to_flush = set()
    rset = cnx.execute('Any X, T WHERE X is CeleryTask, X task_id T')
    pb = ProgressBar(len(rset))
    for eid, task_id in rset:
        entity = cnx.entity_from_eid(eid)
        logs = get_task_logs(task_id)
        if logs is not None:
            to_flush.add(task_id)
        else:
            if entity.task_logs is not None:
                logs = entity.task_logs.read()
        if logs is not None:
            fname = get_log_filename(task_id)
            with gzip.open(fname, 'wb') as f:
                f.write(logs)
        pb.update()
    cnx.commit()
    for task_id in to_flush:
        flush_task_logs(task_id)



# -*- coding: utf-8 -*-
# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""cubicweb-celerytask entity's classes"""
import six

import celery
from celery.result import AsyncResult, result_from_tuple

from cubicweb import NoResultError
from cubicweb.entities import AnyEntity, fetch_config
from cubicweb.view import EntityAdapter
from cubicweb.predicates import is_instance
from cubicweb.server.hook import DataOperationMixIn, Operation

from cw_celerytask_helpers.filelogger import get_task_logs

from cubicweb_celerytask import STATES, FINAL_STATES

_ = six.text_type

_TEST_TASKS = {}
UNKNOWN_TASK_NAME = six.text_type('<unknown>')


def get_tasks():
    """Return tasks to be run (for use in cubicweb test mode)"""
    return _TEST_TASKS.copy()


def run_all_tasks(cnx):
    """Run all pending tasks (for use in cubicweb test mode)"""
    results = {}
    # run all tasks and gather results.
    # Tasks can create other tasks, so run them until there is no one left.
    while _TEST_TASKS:
        task_eid = list(_TEST_TASKS)[0]
        task = _TEST_TASKS.pop(task_eid)
        # Ensure current task id is in the scope of the current test
        if task.id is not None and not cnx.execute(
            'Any X WHERE X is CeleryTask, X task_id %(task_id)s',
            {'task_id': task.freeze().id}
        ):
            continue
        results[task_eid] = task.delay()

    if celery.current_app.conf.task_always_eager:
        for task_eid, result in results.items():
            wf = cnx.entity_from_eid(task_eid).cw_adapt_to('IWorkflowable')
            transition = {
                STATES.SUCCESS: 'finish',
                STATES.FAILURE: 'fail',
            }[result.state]
            comment = result.traceback
            if comment is not None and not isinstance(comment, six.text_type):
                comment = comment.decode('utf-8')
            wf.fire_transition(transition, comment)
    return results


def sync_task_state(cnx, task_id, task_name):
    log = CeleryTaskAdapter
    task_id = six.text_type(task_id)
    result = AsyncResult(task_id)
    if result.state == 'PENDING':
        log.info('Task %s state is unknown', task_id)
        return
    try:
        task = cnx.find('CeleryTask', task_id=task_id).one()
    except NoResultError:
        task = cnx.create_entity('CeleryTask', task_id=task_id,
                                 task_name=task_name or UNKNOWN_TASK_NAME)
        log.info('Created <CeleryTask %s (task_id %s)>',
                 task.eid, task_id)
    task.cw_adapt_to('ICeleryTask').sync_state(task_id, task_name)


def start_async_task(cnx, task, *args, **kwargs):
    """Create and start a new task

    `task` can be either a task name, a task object or a task signature
    """
    task_name = six.text_type(celery.signature(task).task)
    entity = cnx.create_entity('CeleryTask', task_name=task_name)
    entity.cw_adapt_to('ICeleryTask').start(task, *args, **kwargs)
    return entity


def task_in_backend(task_id):
    app = celery.current_app
    if app.conf.task_always_eager:
        return False
    else:
        backend = app.backend
        return backend.get(backend.get_key_for_task(task_id)) is not None


class StartCeleryTaskOp(DataOperationMixIn, Operation):

    def postcommit_event(self):
        global _TEST_TASKS
        if self.cnx.vreg.config.mode == 'test':
            # In test mode, task should run explicitly with run_all_tasks()
            _TEST_TASKS.update(self.cnx.transaction_data.get('celerytask', {}))
        else:
            for eid in self.get_data():
                task = self.cnx.transaction_data.get('celerytask', {}).get(eid)
                if task is not None:
                    task.delay()


class CeleryTask(AnyEntity):
    __regid__ = 'CeleryTask'
    fetch_attrs, cw_fetch_order = fetch_config(('task_name',))

    def dc_title(self):
        return self.task_name

    def dc_long_title(self):
        adapted = self.cw_adapt_to('ICeleryTask')
        state, finished = adapted.state, adapted.finished
        title = self.task_name or self._cw._('subtask')
        if finished:
            title = '%s (%s)' % (title, self._cw._(state))
        return title

    @property
    def progress(self):
        yield self.cw_adapt_to('ICeleryTask').progress
        for subtask in self.reverse_parent_task:
            yield subtask.progress

    @property
    def parent_tasks(self):
        yield self
        for task in self.parent_task:
            for ptask in task.parent_tasks:
                yield ptask

    def child_tasks(self):
        yield self
        for task in self.reverse_parent_task:
            for ctask in task.child_tasks():
                yield ctask


class ICeleryTask(EntityAdapter):
    __regid__ = 'ICeleryTask'
    __abstract__ = True

    def start(self, name, *args, **kwargs):
        eid = self.entity.eid
        task = self.get_task(name, *args, **kwargs)
        self._cw.transaction_data.setdefault('celerytask', {})[eid] = task
        StartCeleryTaskOp.get_instance(self._cw).add_data(eid)

    def get_task(self, name, *args, **kwargs):
        """Should return a celery task / signature or None

        This method is run in a precommit event
        """
        return celery.signature(name, args=args, kwargs=kwargs)

    def sync_state(self, task_id, task_name):
        """Triggered by celery-monitor"""
        raise NotImplementedError

    @property
    def task_id(self):
        raise NotImplementedError

    @property
    def task_name(self):
        raise NotImplementedError

    def revoke(self, terminate=True, signal='SIGKILL'):
        return celery.task.control.revoke(
            [self.task_id], terminate=terminate, signal=signal)

    @property
    def logs(self):
        return get_task_logs(self.task_id) or b''

    @property
    def result(self):
        return AsyncResult(self.task_id)

    @property
    def progress(self):
        if celery.current_app.conf.task_always_eager:
            return 1.
        result = self.result
        try:
            if result.info and 'progress' in result.info:
                return result.info['progress']
        except TypeError:
            pass
        if self.entity.reverse_parent_task:
            children = self.entity.reverse_parent_task
            return sum(child.cw_adapt_to('ICeleryTask').progress
                       for child in children) / len(children)
        if result.state == STATES.SUCCESS:
            return 1.
        return 0.

    @property
    def state(self):
        return self.result.state

    @property
    def finished(self):
        return self.state in FINAL_STATES


class CeleryTaskAdapter(ICeleryTask):
    """Base adapter that store task call args in the transaction"""

    __select__ = ICeleryTask.__select__ & is_instance('CeleryTask')

    def attach_task(self, task, seen, parent=None):
        task_id = six.text_type(task.freeze().id)
        if parent is None:
            parent = self.entity
        if self.entity.task_id is None:
            self.entity.cw_set(task_id=task_id)
        elif task_id not in seen:
            task_name = six.text_type(task.task)
            parent = self._cw.create_entity('CeleryTask',
                                            task_id=six.text_type(task_id),
                                            task_name=task_name,
                                            parent_task=parent)
        seen.add(task_id)
        if task.name in ('celery.chain', 'celery.group'):
            for subtask in task.tasks:
                self.attach_task(subtask, seen, parent)
        if task.name == 'celery.chord':
            self.attach_task(task.body, seen, parent)
            for subtask in task.tasks.tasks:
                self.attach_task(subtask, seen, parent)

    def get_task(self, name, *args, **kwargs):
        task = super(CeleryTaskAdapter, self).get_task(
            name, *args, **kwargs)
        self.attach_task(task, set())
        return task

    @property
    def task_id(self):
        return self.entity.task_id

    @property
    def task_name(self):
        return self.entity.task_name

    def revoke(self, terminate=True, signal='SIGKILL'):
        to_revoke = set([e.task_id for e in self.entity.child_tasks()])
        return celery.task.control.revoke(
            list(to_revoke), terminate=terminate, signal=signal)

    def attach_result(self, result):
        def tree(result, seen=None):
            if seen is None:
                seen = set()
            if result.parent:
                for r in tree(result.parent, seen):
                    yield r
            for child in result.children or []:
                for r in tree(child, seen):
                    yield r

            if isinstance(result, AsyncResult):
                rresult = result.result
                if (isinstance(rresult, dict)
                        and "celerytask_subtasks" in rresult):
                    subtasks = result_from_tuple(
                        rresult["celerytask_subtasks"])
                    for r in tree(subtasks, seen):
                        yield r

                if result.task_id not in seen:
                    seen.add(result.task_id)
                yield result

        for asr in tree(result):
            task_id = six.text_type(asr.id)
            try:
                cwtask = self._cw.find('CeleryTask', task_id=task_id).one()
            except NoResultError:
                cwtask = self._cw.create_entity(
                    'CeleryTask',
                    task_name=UNKNOWN_TASK_NAME,
                    task_id=six.text_type(task_id))
                self.info("Create <CeleryTask %s (task_id %s)>",
                          cwtask.eid, task_id)
            if not cwtask.parent_task and self.entity is not cwtask:
                self.info('Set %s parent_task to %s (%s)', cwtask.task_id,
                          self.entity, self.entity.task_id)
                cwtask.cw_set(parent_task=self.entity)

    def sync_state(self, task_id, task_name, commit=True):
        if (self.entity.task_name == UNKNOWN_TASK_NAME
                and task_name is not None):
            self.info('Update <CeleryTask %s (task_id %s)> name to %s',
                      self.entity.eid, task_id, task_name)
            self.entity.cw_set(task_name=six.text_type(task_name))

        result = self.result
        if result.ready():
            self.attach_result(result)

        transition = {
            STATES.SUCCESS: 'finish',
            STATES.FAILURE: 'fail',
            STATES.STARTED: 'start',
            STATES.REVOKED: 'fail',
            'PROGRESS': 'start',
        }.get(result.state)
        if transition is not None:
            self.info('<CeleryTask %s (task_id %s)> %s', self.entity.eid,
                      task_id, transition)
            wf = self.entity.cw_adapt_to('IWorkflowable')
            if (result.traceback is None and result.state == STATES.FAILURE
                    and result.result is not None):
                comment = six.text_type(result.result)
            else:
                comment = result.traceback
            wf.fire_transition_if_possible(transition, comment)
        else:
            self.info('<CeleryTask %s (task_id %s)> no transition found for '
                      'state %s', self.entity.eid, task_id, result.state)

        if commit:
            self._cw.commit()

    @property
    def state(self):
        db_state = self.entity.cw_adapt_to('IWorkflowable').state
        db_final_state_map = {'done': STATES.SUCCESS, 'failed': STATES.FAILURE}
        if db_state in db_final_state_map:
            return db_final_state_map[db_state]
        elif task_in_backend(self.task_id):
            return super(CeleryTaskAdapter, self).state
        return _('unknown state')
